﻿using System.Collections.Generic;
using SimpleLang.Visitors;

namespace ProgramTree
{
    public enum AssignType { Assign, AssignPlus, AssignMinus, AssignMult, AssignDivide };

    public abstract class Node // базовый класс для всех узлов    
    {
        public abstract void Visit(Visitor v);
    }

    public abstract class ExprNode : Node // базовый класс для всех выражений
    {
    }

    public class IdNode : ExprNode
    {
        public string Name { get; set; }
        public IdNode(string name) { Name = name; }
        public override void Visit(Visitor v)
        {
            if (v.GetType() == typeof(CommonlyUsedVarVisitor))
            {
                if (((CommonlyUsedVarVisitor)v).Dict.ContainsKey(Name))
                    ((CommonlyUsedVarVisitor)v).Dict[Name]++;
                else
                    ((CommonlyUsedVarVisitor)v).Dict[Name] = 1;
            }
            else if(v.GetType() == typeof(ChangeVarIdVisitor))
            {
                if (Name == ((ChangeVarIdVisitor)v).get_from())
                    Name = ((ChangeVarIdVisitor)v).get_to();
            }
            v.VisitIdNode(this);
        }
    }

    public class IntNumNode : ExprNode
    {
        public int Num { get; set; }
        public IntNumNode(int num) { Num = num; }
        public override void Visit(Visitor v)
        {
            v.VisitIntNumNode(this);
        }
    }

    public class BinOpNode : ExprNode
    {
        public ExprNode Left { get; set; }
        public ExprNode Right { get; set; }
        public char Op { get; set; }
        public BinOpNode(ExprNode Left, ExprNode Right, char Op) 
        {
            this.Left = Left;
            this.Right = Right;
            this.Op = Op;
        }
        public override void Visit(Visitor v)
        {
            if (v.GetType() == typeof(ExprComplexityVisitor))
            {
                if (Op == '+' || Op == '-')
                    ((ExprComplexityVisitor)v).ExprComplexity[((ExprComplexityVisitor)v).ExprComplexity.Count - 1] += 1;
                else
                    if (Op == '*' || Op == '/')
                        ((ExprComplexityVisitor)v).ExprComplexity[((ExprComplexityVisitor)v).ExprComplexity.Count - 1] += 3;
            }
            v.VisitBinOpNode(this);
        }
    }

    public abstract class StatementNode : Node // базовый класс для всех операторов
    {
    }

    public class AssignNode : StatementNode
    {
        public IdNode Id { get; set; }
        public ExprNode Expr { get; set; }
        public AssignType AssOp { get; set; }
        public AssignNode(IdNode id, ExprNode expr, AssignType assop = AssignType.Assign)
        {
            Id = id;
            Expr = expr;
            AssOp = assop;
        }
        public override void Visit(Visitor v)
        {
            if (v.GetType() == typeof(CountCyclesOpVisitor))
            {
                if (((CountCyclesOpVisitor)v).InCycle)
                    ((CountCyclesOpVisitor)v).OpCnt++;
            }
            else if (v.GetType() == typeof(ExprComplexityVisitor))
                ((ExprComplexityVisitor)v).ExprComplexity.Add(0);
            v.VisitAssignNode(this);
        }
    }

    public class CycleNode : StatementNode
    {
        public ExprNode Expr { get; set; }
        public StatementNode Stat { get; set; }
        public CycleNode(ExprNode expr, StatementNode stat)
        {
            Expr = expr;
            Stat = stat;
        }
        public override void Visit(Visitor v)
        {
            if (v.GetType() == typeof(CountCyclesOpVisitor))
            {
                ((CountCyclesOpVisitor)v).CyclesCnt++;
                ((CountCyclesOpVisitor)v).InCycle = true;
            }
            else if (v.GetType() == typeof(ExprComplexityVisitor))
                ((ExprComplexityVisitor)v).ExprComplexity.Add(0);
            else if (v.GetType() == typeof(MaxNestCyclesVisitor))
            {
                if(Stat.GetType() == typeof(CycleNode))
                    ((MaxNestCyclesVisitor)v).MaxNest++;  
            }
            else if(v.GetType() == typeof(MaxIfCycleNestVisitor))
                ((MaxIfCycleNestVisitor)v).MaxNest++;
            v.VisitCycleNode(this);
            if (v.GetType() == typeof(CountCyclesOpVisitor))
                ((CountCyclesOpVisitor)v).InCycle = false;
        }
    }

    public class BlockNode : StatementNode
    {
        public List<StatementNode> StList = new List<StatementNode>();
        public BlockNode(StatementNode stat)
        {
            Add(stat);
        }
        public void Add(StatementNode stat)
        {
            StList.Add(stat);
        }
        public override void Visit(Visitor v)
        {
            if (v.GetType() == typeof(MaxNestCyclesVisitor))
            {
                if(StList[StList.Count-1].GetType()==typeof(CycleNode))
                    ((MaxNestCyclesVisitor)v).MaxNest++;
            }
                v.VisitBlockNode(this);
        }
    }

    public class WriteNode : StatementNode
    {
        public ExprNode Expr { get; set; }
        public WriteNode(ExprNode Expr)
        {
            this.Expr = Expr;
        }
        public override void Visit(Visitor v)
        {
            if (v.GetType() == typeof(ExprComplexityVisitor))
                ((ExprComplexityVisitor)v).ExprComplexity.Add(0);
            v.VisitWriteNode(this);
        }
    }

    public class EmptyNode : StatementNode
    {
        public override void Visit(Visitor v)
        {
            v.VisitEmptyNode(this);
        }
    }

    public class VarDefNode : StatementNode
    {
        public List<IdNode> vars = new List<IdNode>();
        public VarDefNode(IdNode id)
        {
            Add(id);
        }

        public void Add(IdNode id)
        {
            vars.Add(id);
        }
        public override void Visit(Visitor v)
        {
            v.VisitVarDefNode(this);
        }
    }

    public class IfNode : StatementNode
    {
        public ExprNode Expr { get; set; }
        public StatementNode Stat1 { get; set; }
        public StatementNode Stat2 { get; set; }
        public IfNode(ExprNode expr, StatementNode stat1)
        {
            Expr = expr;
            Stat1 = stat1;
        }
        public IfNode(ExprNode expr, StatementNode stat1, StatementNode stat2)
        {
            Expr = expr;
            Stat1 = stat1;
            Stat2 = stat2;
        }
        public override void Visit(Visitor v)
        {
            if (v.GetType() == typeof(MaxIfCycleNestVisitor))
                ((MaxIfCycleNestVisitor)v).MaxNest++;
            v.VisitIfNode(this);
        }
    }
}